"""gustaf/gustaf/io/mfem.py.

io functions for mfem. Supports simple linear elements (straight meshes)
For detailed information, see: https://mfem.org/mesh-
format-v1.0/#straight-meshes
"""

import numpy as np

from gustaf import settings
from gustaf.faces import Faces
from gustaf.volumes import Volumes

geometry_types = {
    "POINT": 0,
    "SEGMENT": 1,
    "TRIANGLE": 2,
    "SQUARE": 3,
    "TETRAHEDRON": 4,
    "CUBE": 5,
    "PRISM": 6,
}


def load(fname):
    """Load mesh in MFEM format. Loads vertices and their connectivity.
    Currently cannot process boundary.

    Parameters
    ------------
    fname: str

    Returns
    ------------
    mesh
    """

    def extract_values(fname, start_index, n_lines, total_lines, dtype):
        """Extract information from file. Reads [n_lines] lines from file
        [fname] starting at line [start_index] and returns array of type
        [dtype].

        Parameters
        ------------
        fname: str
        start_index: int
        n_lines: int
        total_lines: int
            Number of total lines of file
        dtype: (NumPy) data type

        Returns
        ------------
        NumPy Array
        """
        end_index = total_lines - (start_index + n_lines + 2)
        return np.genfromtxt(
            fname,
            delimiter=" ",
            skip_header=start_index,
            skip_footer=end_index,
            dtype=dtype,
        )

    with open(fname) as f:
        lines = f.readlines()
        total_lines = len(lines)

        # Read values of keywords
        keywords = ["dimension", "elements", "boundary", "vertices"]
        indices = [lines.index(f"{keyword}\n") for keyword in keywords]
        dimension, n_elements, n_boundaries, n_vertices = (
            int(lines[index + 1]) for index in indices
        )
        vdim = int(lines[indices[-1] + 2])
        # Extract values
        elements = extract_values(
            fname, indices[1] + 2, n_elements, total_lines, settings.INT_DTYPE
        )
        if elements.shape[0] != n_elements:
            raise ValueError("Number of elements do not match.")
        boundary = extract_values(
            fname,
            indices[2] + 2,
            n_boundaries,
            total_lines,
            settings.INT_DTYPE,
        )
        if boundary.shape[0] != n_boundaries:
            raise ValueError("Number of boundaries do not match.")
        vertices = extract_values(
            fname,
            indices[3] + 3,
            n_vertices,
            total_lines,
            settings.FLOAT_DTYPE,
        )
        if vertices.shape != (n_vertices, vdim):
            raise ValueError("Number of vertices do not match.")

        connectivity = elements[:, 2:]
        if dimension == 2:
            mesh = Faces(vertices=vertices, faces=connectivity)
        elif dimension == 3:
            mesh = Volumes(vertices=vertices, volumes=connectivity)

        return mesh


def format_array(array):
    """Format NumPy array into string. Each entry in a row is separated by
    a blank space and every row is separated by a new line.

    Parameters
    ------------
    array: NumPy array

    Returns
    ------------
    str
    """
    row_strings = [" ".join(map(str, row)) for row in array]
    return "\n".join(row_strings)


def export(fname, mesh):
    """Export mesh in MFEM format. Supports 2D triangle and quadrilateral
    meshes as well as 3D tetrahedral and hexahedral meshes. Also extracts
    boundary information from the GustafObject.BC variable. Does not
    support different element attributes or difference in vertex dimension
    and mesh dimension.

    Parameters
    ------------
    fname: str
    mesh: Faces, Volumes

    Returns
    ------------
    None
    """
    # Get dimension info based on vertices and choose corresponding export
    dim = mesh.vertices.shape[1]

    if dim == 2:
        elements_string, boundary_string, vertices_string = _export_2d(mesh)
    elif dim == 3:
        elements_string, boundary_string, vertices_string = _export_3d(mesh)
    else:
        raise NotImplementedError(
            f"Sorry, we cannot export meshes of dimension {dim}."
        )

    with open(fname, "w") as f:
        f.write("MFEM mesh v1.0\n\n")
        f.write(f"dimension\n{dim}\n\n")
        f.write(elements_string)
        f.write(boundary_string)
        f.write(vertices_string)


def _export_2d(mesh):
    """Internal function to export 2D meshes.

    Parameters
    ------------
    mesh: Faces

    Returns
    ------------
    elements_string, boundary_string, vertices_string
    """
    element_attribute = 1  # Other numbers not yet supported
    elements = mesh.elements
    n_elements = elements.shape[0]
    if mesh.whatami == "tri":
        geometry_type = geometry_types["TRIANGLE"]
    elif mesh.whatami == "quad":
        geometry_type = geometry_types["SQUARE"]
    else:
        raise NotImplementedError(
            f"Sorry, we cannot export 2D meshes of type {mesh.whatami}."
        )
    e = np.ones((n_elements, 1), dtype=settings.INT_DTYPE)
    elements_array = np.hstack(
        (element_attribute * e, geometry_type * e, elements)
    )
    elements_array_string = format_array(elements_array)
    elements_string = f"elements\n{n_elements}\n"
    elements_string += f"{elements_array_string}\n\n"

    # Boundary
    edges = mesh.edges()
    nboundary_edges = sum(map(len, mesh.BC.values()))
    boundary_array = np.empty((nboundary_edges, 4), dtype=settings.INT_DTYPE)
    startrow = 0
    # Add boundary one by one as SEGMENTs
    for bid, edgeids in mesh.BC.items():
        nedges = len(edgeids)
        e = np.ones(nedges).reshape(-1, 1)
        vertex_list = edges[edgeids, :]
        boundary_array[startrow : (startrow + nedges), :] = np.hstack(
            (int(bid) * e, geometry_types["SEGMENT"] * e, vertex_list)
        )
        startrow += nedges

    boundary_array_string = format_array(boundary_array)
    boundary_string = f"boundary\n{nboundary_edges}\n"
    boundary_string += f"{boundary_array_string}\n\n"

    # Vertices
    vdim = 2  # Currently only option
    vertices_array_string = format_array(mesh.vertices)
    nvertices = mesh.vertices.shape[0]
    vertices_string = f"vertices\n{nvertices}\n{vdim}\n"
    vertices_string += f"{vertices_array_string}"
    return elements_string, boundary_string, vertices_string


def _export_3d(mesh):
    """Internal function to export 3D meshes.

    Parameters
    ------------
    mesh: Volumes

    Returns
    ------------
    elements_string, boundary_string, vertices_string
    """
    element_attribute = 1  # Other numbers not yet supported
    elements = mesh.elements
    n_elements = elements.shape[0]
    if mesh.whatami == "tet":
        body_geometry_type = geometry_types["TETRAHEDRON"]
        face_geometry_type = geometry_types["TRIANGLE"]
        n_face_vertices = 3
    elif mesh.whatami == "hexa":
        body_geometry_type = geometry_types["CUBE"]
        face_geometry_type = geometry_types["SQUARE"]
        n_face_vertices = 4
    else:
        raise NotImplementedError(
            f"Sorry, we cannot export 3D meshes of type {mesh.whatami}."
        )
    e = np.ones((n_elements, 1), dtype=settings.INT_DTYPE)
    elements_array = np.hstack(
        (element_attribute * e, body_geometry_type * e, elements)
    )
    elements_array_string = format_array(elements_array)
    elements_string = f"elements\n{n_elements}\n"
    elements_string += f"{elements_array_string}\n\n"

    # Boundary
    faces = mesh.faces()

    nboundary_faces = sum(map(len, mesh.BC.values()))
    boundary_array = np.empty(
        (nboundary_faces, n_face_vertices + 2), dtype=settings.INT_DTYPE
    )
    startrow = 0
    # Add boundary one by one as TRIANGLEs
    for bid, faceids in mesh.BC.items():
        nfaces = len(faceids)
        e = np.ones(nfaces).reshape(-1, 1)
        vertex_list = faces[faceids, :]
        boundary_array[startrow : (startrow + nfaces), :] = np.hstack(
            (int(bid) * e, face_geometry_type * e, vertex_list)
        )
        startrow += nfaces

    boundary_array_string = format_array(boundary_array)
    boundary_string = f"boundary\n{nboundary_faces}\n"
    boundary_string += f"{boundary_array_string}\n\n"

    # Vertices
    vdim = 3  # Currently only option
    vertices_array_string = format_array(mesh.vertices)
    nvertices = mesh.vertices.shape[0]
    vertices_string = f"vertices\n{nvertices}\n{vdim}\n"
    vertices_string += f"{vertices_array_string}"
    return elements_string, boundary_string, vertices_string



"""gustaf/gustaf/volumes.py."""

import numpy as np

from gustaf import helpers, settings, show, utils
from gustaf.faces import Faces
from gustaf.helpers.options import Option


class VolumesShowOption(helpers.options.ShowOption):
    """
    Show options for vertices.
    """

    _valid_options = helpers.options.make_valid_options(
        *helpers.options.vedo_common_options,
        Option("vedo", "lw", "Width of edges (lines) in pixel units.", (int,)),
        Option(
            "vedo", "lc", "Color of edges (lines).", (int, str, tuple, list)
        ),
    )

    _helps = "Volumes"

    def _initialize_showable(self):
        """
        Initialize volumes as vedo.UGrid or visually equivalent vedo.Mesh

        Parameters
        ----------
        None

        Returns
        -------
        volumes: vedo.UGrid or vedo.Mesh
        """
        # without a data to plot on the surface, return vedo.UGrid
        # if vertex_ids is on, we will go with mesh
        if (
            self.get("data", None) is None
            and not self.get("vertex_ids", False)
            and not self.get("arrow_data", False)
            and not show.is_ipython
        ):
            from vtk import VTK_HEXAHEDRON as herr_hexa
            from vtk import VTK_TETRA as frau_tetra

            to_vtktype = {"tet": frau_tetra, "hexa": herr_hexa}
            grid_type = to_vtktype[self._helpee.whatami]
            u_grid = show.vedoUGrid(
                [
                    self._helpee.const_vertices,
                    self._helpee.const_volumes,
                    [grid_type] * len(self._helpee.const_volumes),
                ]
            )

            for option in ["lw", "lc"]:
                val = self.get(option, False)
                if val:
                    getattr(u_grid, option)(val)

            return u_grid.c("hotpink")

        # to show data, let's use Faces. This will plot all the elements
        # as well as invisible ones. This will at least try to avoid
        # duplicating faces.  If you wanna see inside faces, try
        # as_shrunk_faces = volumes.to_faces(unique=False).shrink(.8)
        faces = self._helpee.to_faces(unique=True)
        self.copy_valid_options(faces.show_options)

        return faces.show_options._initialize_showable()


class Volumes(Faces):
    kind = "volume"

    const_faces = helpers.raise_if.invalid_inherited_attr(
        "Faces.const_faces",
        __qualname__,
        property_=True,
    )
    update_faces = helpers.raise_if.invalid_inherited_attr(
        "Faces.update_edges",
        __qualname__,
        property_=False,
    )

    __slots__ = (
        "_volumes",
        "_const_volumes",
    )

    __show_option__ = VolumesShowOption
    __boundary_class__ = Faces

    def __init__(
        self,
        vertices=None,
        volumes=None,
        elements=None,
    ):
        """Volumes. It has vertices and volumes. Volumes could be tetrahedrons
        or hexahedrons.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        volumes: (n, 4) or (n, 8) np.ndarray
        """
        super().__init__(vertices=vertices)
        if volumes is not None:
            self.volumes = volumes
        elif elements is not None:
            self.volumes = elements

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def faces(self):
        """Faces here aren't main property. So this needs to be computed.

        Parameters
        -----------
        None

        Returns
        --------
        faces: (n, 3) or (n, 4) np.ndarray
        """
        whatami = self.whatami
        faces = None
        if whatami.startswith("tet"):
            faces = utils.connec.tet_to_tri(self.volumes)
        elif whatami.startswith("hexa"):
            faces = utils.connec.hexa_to_quad(self.volumes)

        return faces

    @classmethod
    def whatareyou(cls, volume_obj):
        """overwrites Faces.whatareyou to tell you is this volume is tet or
        hexa.

        Parameters
        -----------
        volume_obj: Volumes

        Returns
        --------
        whatareyou: str
        """
        if not cls.kind.startswith(volume_obj.kind):
            raise TypeError("Given obj is not {cls.__qualname__}")

        if volume_obj.volumes.shape[1] == 4:
            return "tet"

        elif volume_obj.volumes.shape[1] == 8:
            return "hexa"

        else:
            raise ValueError(
                "Invalid volumes connectivity shape. It should be (n, 4) "
                f"or (n, 8), but given: {volume_obj.volumes.shape}"
            )

    @property
    def volumes(self):
        """Returns volumes.

        Parameters
        -----------
        None

        Returns
        --------
        volumes: (n, 4) or (n, 8) np.ndarray
        """
        return self._volumes

    @volumes.setter
    def volumes(self, vols):
        """volumes setter. Similar to vertices, this will be a tracked array.

        Parameters
        -----------
        vols: (n, 4) or (n, 8) np.ndarray

        Returns
        --------
        None
        """
        self._volumes = helpers.data.make_tracked_array(
            vols,
            settings.INT_DTYPE,
            copy=False,
        )
        if vols is not None:
            utils.arr.is_one_of_shapes(
                vols,
                ((-1, 4), (-1, 8)),
                strict=True,
            )
        # same, but non-writeable view of tracked array
        self._const_volumes = self._volumes.view()
        self._const_volumes.flags.writeable = False

    @property
    def const_volumes(self):
        """Returns non-writeable view of volumes.

        Parameters
        -----------
        None

        Returns
        --------
        const_volumes: (n, 4) or (n, 8) np.ndarray
        """
        return self._const_volumes

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def sorted_volumes(self):
        """Sort volumes along axis=1.

        Parameters
        -----------
        None

        Returns
        --------
        volumes_sorted: (volumes.shape) np.ndarray
        """
        volumes = self._get_attr("volumes")

        return np.sort(volumes, axis=1)

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def unique_volumes(self):
        """Returns a namedtuple of unique volumes info. Similar to
        unique_edges.

        Parameters
        -----------
        None

        Returns
        --------
        unique_info: Unique2DIntegers
          valid attributes are {values, ids, inverse, counts}
        """
        unique_info = utils.connec.sorted_unique(
            self.sorted_volumes(),
            sorted_=True,
        )

        volumes = self._get_attr("volumes")

        unique_info.values[:] = volumes[unique_info.ids]

        return unique_info

    def update_volumes(self, *args, **kwargs):
        """Alias to update_elements."""
        self.update_elements(*args, **kwargs)

    def to_faces(self, unique=True):
        """Returns Faces obj.

        Parameters
        -----------
        unique: bool
          Default is True. If True, only takes unique faces.

        Returns
        --------
        faces: Faces
        """
        return Faces(
            self.vertices,
            faces=self.unique_faces().values if unique else self.faces(),
        )



"""gustaf/gustaf/faces.py."""

import numpy as np
from packaging.version import Version

from gustaf import helpers, settings, show, utils
from gustaf.edges import Edges
from gustaf.helpers.options import Option

# special types for face texture option
try:
    import vedo

    if Version(vedo.__version__) < Version("2025.5.3"):
        vedoPicture = vedo.Picture
    else:
        vedoPicture = vedo.Image
    # there are other ways to get here, but this is exact path for our use
    vtkTexture = vedo.vtkclasses.vtkTexture
except ImportError as err:
    vedoPicture = helpers.raise_if.ModuleImportRaiser("vedo", err)
    vtkTexture = helpers.raise_if.ModuleImportRaiser("vedo", err)


class FacesShowOption(helpers.options.ShowOption):
    """
    Show options for vertices.
    """

    _valid_options = helpers.options.make_valid_options(
        *helpers.options.vedo_common_options,
        Option("vedo", "lw", "Width of edges (lines) in pixel units.", (int,)),
        Option(
            "vedo", "lc", "Color of edges (lines).", (int, str, tuple, list)
        ),
        Option(
            "vedo",
            "texture",
            "Texture of faces in array, vedo.Picture/Image, vtk.vtkTexture, "
            "or path to an image.",
            (np.ndarray, tuple, list, str, vedoPicture, vtkTexture),
        ),
    )

    _helps = "Faces"

    def _initialize_showable(self):
        """
        Initializes Faces as vedo.Mesh

        Parameters
        ----------
        None

        Returns
        -------
        faces: vedo.Mesh
        """

        faces = show.vedo.Mesh(
            [self._helpee.const_vertices, self._helpee.const_faces],
        )

        for option in ["lw", "lc", "texture"]:
            val = self.get(option, False)
            if val:
                getattr(faces, option)(val)

        return faces


class Faces(Edges):
    kind = "face"

    const_edges = helpers.raise_if.invalid_inherited_attr(
        "Edges.const_edges",
        __qualname__,
        property_=True,
    )
    update_edges = helpers.raise_if.invalid_inherited_attr(
        "Edges.update_edges",
        __qualname__,
        property_=False,
    )
    dashed = helpers.raise_if.invalid_inherited_attr(
        "Edges.dashed",
        __qualname__,
        property_=False,
    )

    __slots__ = (
        "_faces",
        "_const_faces",
        "BC",
    )

    __show_option__ = FacesShowOption
    __boundary_class__ = Edges

    def __init__(
        self,
        vertices=None,
        faces=None,
        elements=None,
    ):
        """Faces. It has vertices and faces. Faces could be triangles or
        quadrilaterals.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        faces: (n, 3) or (n, 4) np.ndarray
        """
        super().__init__(vertices=vertices)
        if faces is not None:
            self.faces = faces

        elif elements is not None:
            self.faces = elements

        self.BC = {}

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def edges(self):
        """Edges from here aren't main property. So this needs to be computed.

        Parameters
        -----------
        None

        Returns
        --------
        edges: (n, 2) np.ndarray
        """
        self._logd("computing edges")
        faces = self._get_attr("faces")

        return utils.connec.faces_to_edges(faces)

    @property
    def whatami(self):
        """Determines whatami.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        return type(self).whatareyou(self)

    @classmethod
    def whatareyou(cls, face_obj):
        """classmethod that tells you if the Faces is tri or quad or invalid
        kind.

        Parameters
        -----------
        face_obj: Faces

        Returns
        --------
        whatareyou: str
        """
        if not cls.kind.startswith(face_obj.kind):
            raise TypeError("Given obj is not {cls.__qualname__}")

        if face_obj.faces.shape[1] == 3:
            return "tri"

        elif face_obj.faces.shape[1] == 4:
            return "quad"

        else:
            raise ValueError(
                "Invalid faces connectivity shape. It should be (n, 3) or "
                f"(n, 4), but given: {face_obj.faces.shape}"
            )

    @property
    def faces(self):
        """Returns faces.

        Parameters
        -----------
        None

        Returns
        --------
        faces
        """
        self._logd("returning faces")
        return self._faces

    @faces.setter
    def faces(self, fs):
        """Faces setter. Similar to vertices, this will be a tracked array.

        Parameters
        -----------
        fs: (n, 2) np.ndarray

        Returns
        --------
        None
        """
        self._logd("setting faces")

        self._faces = helpers.data.make_tracked_array(
            fs,
            settings.INT_DTYPE,
            copy=False,
        )
        # shape check
        if fs is not None:
            utils.arr.is_one_of_shapes(
                fs,
                ((-1, 3), (-1, 4)),
                strict=True,
            )

        # same, but non-writeable view of tracked array
        self._const_faces = self._faces.view()
        self._const_faces.flags.writeable = False

    @property
    def const_faces(self):
        """Returns non-writeable view of faces.

        Parameters
        -----------
        None

        Returns
        --------
        const_faces: (n, 2
        """
        return self._const_faces

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def sorted_faces(self):
        """Similar to edges_sorted but for faces.

        Parameters
        -----------
        None

        Returns
        --------
        sorted_faces: (self.faces.shape) np.ndarray
        """
        faces = self._get_attr("faces")

        return np.sort(faces, axis=1)

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def unique_faces(self):
        """Returns a namedtuple of unique faces info. Similar to unique_edges.

        Parameters
        -----------
        None

        Returns
        --------
        unique_info: Unique2DIntegers
          valid attributes are {values, ids, inverse, counts}
        """
        unique_info = utils.connec.sorted_unique(
            self.sorted_faces(), sorted_=True
        )

        faces = self._get_attr("faces")

        unique_info.values[:] = faces[unique_info.ids]

        return unique_info

    @helpers.data.ComputedMeshData.depends_on(["elements"])
    def single_faces(self):
        """Returns indices of very unique faces: faces that appear only once.
        For well constructed volumes, this can be considered as surfaces.

        Parameters
        -----------
        None

        Returns
        --------
        single_faces: (m,) np.ndarray
        """
        unique_info = self.unique_faces()

        return unique_info.ids[unique_info.counts == 1]

    def update_faces(self, *args, **kwargs):
        """Alias to update_elements."""
        self.update_elements(*args, **kwargs)

    def to_edges(self, unique=True):
        """Returns Edges obj.

        Parameters
        -----------
        unique: bool
          Default is True. If True, only takes unique edges.

        Returns
        --------
        edges: Edges
        """
        return Edges(
            self.vertices,
            edges=self.unique_edges().values if unique else self.edges(),
        )

    def to_subelements(
        self,
        unique=True,  # noqa ARG002 # used inside the return eval str
    ):
        """Returns current elements represented as its boundary element class.
        For faces, this is equivalent to `to_edges()`.
        For volumes, `to_faces()`.

        Parameters
        ----------
        unique: bool
          Default is True. If True, only takes unique edges.

        Returns
        -------
        subelements: boundary class
        """
        return getattr(
            self, f"to_{self.__boundary_class__.__qualname__.lower()}"
        )(unique)



"""Handles the processing of markdown files.

This script is used to process the markdown files in the project that are also
reused in the documentation. The script will replace the relative links in the
markdown files with relative links that are correct when the documentation is
built.

The paths change since the documentation is built from the docs folder and not
from the root of the project.

Author: Clemens Fricke
"""

import os
import pathlib
import re

# Path to this file.
file_path = os.path.abspath(os.path.dirname(__file__))
repo_root = str(pathlib.Path(__file__).resolve()).split("docs")[0]


def get_markdown_links(line: str) -> str:
    """Get the markdown links from a string.

    Args:
        line (str): Text.

    Returns:
        str: Markdown links.
    """
    possible = re.findall(r"\[(.*?)\]\((.*?)\)", line)
    return possible if possible else ""


def get_github_path_from(link):
    """Substitute the path to the github repository.

    This will expand the link to the github repository. This is used to create
    pages that are independent of the documentation. Like for the long
    description on PyPI. Normally we try to use relative links in the files
    to guarantee that the documentation is also available offline. But for
    the long description on PyPI we need to use absolute links to an online
    repository like the github raw files.

    Args:
        link (str): Relative path to the file.

    Returns:
        str: Https path to the file on github.
    """
    return os.path.abspath(link).replace(
        repo_root, "https://raw.githubusercontent.com/tataratat/gustaf/main/"
    )


def get_abs_path_from(path: str, sub_folder: str = "") -> str:
    """Get the absolute path from the projects base directory.

    Args:
        path (str): Relative file path given from the projects base directory.

    Returns:
        str: Absolute path to the given file.
    """
    return os.path.abspath(os.path.join(file_path, "../../", sub_folder, path))


# Folder to save the processed markdown files to.
folder_to_save_to = os.path.abspath(get_abs_path_from("docs/md/"))

# List of markdown files that are used in the documentation.
markdown_files = [
    get_abs_path_from("README.md"),
    get_abs_path_from("CONTRIBUTING.md"),
]


def process_file(
    file: str, relative_links: bool = True, return_content: bool = False
):
    """Process a markdown file.

    Args:
        file (str): Path to the markdown file.
        relative_links (bool, optional):
            Generate relative links. Defaults to False.
        return_content (bool, optional):
            Return the content instead of saving it to a file.
            Defaults to False.

    Returns:
        str:
            Content of the markdown file.
            Only returned if return_content is True.
    """
    # read in the content of the markdown file
    with open(file) as f:
        content = f.read()
    # get all links from the markdown file
    links = get_markdown_links(content)

    for item in links:
        if item[1].startswith(("http", "#")):  # skip http links and anchors
            content = content.replace(
                f"[{item[0]}]({item[1]})",
                f"<a href='{item[1]}'>{item[0]}</a>",
            )
            continue
        elif not relative_links:  # generate links to github repo
            new_path = get_github_path_from(get_abs_path_from(item[1]))
        else:  # generate relative links
            new_path = os.path.relpath(get_abs_path_from(item[1]), file_path)
        content = content.replace(item[1], new_path)

    if return_content:
        return content

    with open(
        os.path.join(folder_to_save_to, os.path.basename(file)), "w"
    ) as f:
        f.write(content)


if __name__ == "__main__":
    # Process all markdown files
    for file in markdown_files:
        process_file(file)



"""gustaf/gustaf/utils/tictoc.py.

Timer that tics, tocs and logs.
"""

from time import perf_counter as now

from gustaf._base import GustafBase


class Tic(GustafBase):
    """
    Timer class for easier time measurement.
    """

    __slots__ = ("_title", "_names", "_laps", "_logger")

    # add short cut to perf_counter in case you just want pure "now"
    now = now

    def __init__(self, title="untitled", log_level="debug"):
        """
        Configures the timer and logger. Marks the starting point.

        Parameters
        ----------
        title: str
          Title for this measurement. Defaults to "untitled".
        log_level: str
          Valid options are {"info", "debug", "warning"}. Defaults to "debug".
        """
        # select logger based on logger level
        self._logger = eval(f"self._log{str(log_level).lower()[0]}")

        # initialize names
        self._names = []

        # give name
        self._title = str(title)

        # start
        self._laps = [now()]

    def toc(self, name=None, log=False):
        """
        Adds now to the measurements.

        Parameters
        ----------
        name: str
          Name of this specific measurement lap.
          By default, it assigns a number to this lap.
        log: bool
          If True, will log lapsed time. Defaults to None.

        Returns
        -------
        None
        """
        # measure now
        self._laps.append(now())

        self._names.append(
            f"lap-{len(self._laps) - 2}" if name is None else str(name)
        )

        if log:
            self._logger(
                f"{self._title} - {name}:  {self._laps[-2] - self._laps[-1]}"
            )

    def summary(self, log=True, print_=False):
        """
        Prints measurement summary to the log.

        Parameters
        ----------
        log: bool
          Logs the summary. Default is True.
        print_: bool
          Prints the summary. Default is False.


        Returns
        -------
        records: tuple
          Lap timings in a tuple consisting of a list of the lap names and
           times for each lap from the timer start (cumulative time).
        """
        start = self._laps[0]
        cumulative = [f"{lap - start:.10f}" for lap in self._laps[1:]]

        if log or print_:
            message = [f"\n+++ {self._title} - time logs +++\n"]

            name_width = int(max([len(n) for n in self._names]))

            # extract info
            diff = [
                f"{l1 - l0:.10f}"
                for l0, l1 in zip(self._laps[:-1], self._laps[1:])
            ]
            diff_width = int(max([len(d) for d in diff]))
            cumulative_width = int(max([len(c) for c in cumulative]))

            message.append(
                f"{'names'.ljust(name_width)} | "
                f"{'diff'.rjust(diff_width)} | "
                f"{'cumulative'.rjust(cumulative_width)}\n"
            )

            for n, d, c in zip(self._names, diff, cumulative):
                this_line = (
                    f"{n.ljust(name_width)} | "
                    f"{d.rjust(diff_width)} | "
                    f"{c.rjust(cumulative_width)}\n"
                )
                message.append(this_line)

            if log:
                self._logger(*message)
            if print_:
                print(*message)  # noqa: T201

        return self._names.copy(), cumulative



"""gustaf/gustaf/helpers/options.py

Classes to help organize options.
"""

from copy import deepcopy

from gustaf.helpers._base import HelperBase
from gustaf.helpers.raise_if import ModuleImportRaiser


class Option:
    """
    Minimal Class to hold each options. Mainly to replace nested dict.

    Parameters
    ----------
    backends: set
      set of strings.
    key: str
    description: str
    allowed_types: set
      set of types
    default: one of allwed_types
      Optional. Default is None
    """

    __slots__ = (
        "backends",
        "key",
        "description",
        "allowed_types",
        "default",
    )

    def __init__(
        self, backends, key, description, allowed_types, default=None
    ):
        """
        Check types
        """
        if isinstance(backends, str):
            self.backends = {backends}
        elif getattr(backends, "__iter__", False):
            self.backends = set(backends)
        else:
            raise TypeError("Invalid backends type")

        if isinstance(key, str):
            self.key = key
        else:
            raise TypeError("Invalid key type")

        if isinstance(description, str):
            self.description = description
        else:
            raise TypeError("Invalid description type")

        if isinstance(allowed_types, (tuple, list, set)):
            self.allowed_types = tuple(allowed_types)
        else:
            raise TypeError("Invalid allowed_types type")

        if default is None or type(default) in self.allowed_types:
            self.default = default
        else:
            raise TypeError(f"{type(default)} is invalid default type")

    def __repr__(self):
        specific = "\n".join(
            [
                "",
                self.key,
                "=" * len(self.key),
                f"backends: {self.backends}",
                "description:",
                f"  {self.description}",
                "allowed_types:",
                f"  {self.allowed_types}",
                super().__repr__(),
                "",
            ]
        )
        return specific


class SetDefault:
    """
    Default setter object. Can use as argument in `make_valid_options`
    """

    __slots__ = ("key", "default")

    def __init__(self, key, default):
        self.key = key
        self.default = default


# predefine recurring options
vedo_common_options = (
    Option(
        "vedo",
        "c",
        "Color in {rgb, RGB, str of (hex, name), int}",
        (str, tuple, list, int),
    ),
    Option("vedo", "alpha", "Transparency in range [0, 1].", (float, int)),
    Option(
        "vedo",
        "data",
        "Name of vertex_data to show. "
        "Object must have vertex_data with the same name.",
        (str,),
    ),
    Option("vedo", "vertex_ids", "Show ids of vertices", (bool,)),
    Option("vedo", "element_ids", "Show ids of elements", (bool,)),
    Option(
        "vedo",
        "lighting",
        "Lighting options {'default', 'metallic', 'plastic', 'shiny', "
        "'glossy', 'ambient', 'off'}",
        (str,),
    ),
    Option("vedo", "cmap", "Colormap for vertex_data plots.", (str,)),
    Option("vedo", "vmin", "Minimum value for cmap", (float, int)),
    Option("vedo", "vmax", "Maximum value for cmap", (float, int)),
    Option(
        "vedo",
        "cmap_alpha",
        "Colormap Transparency in range [0, 1].",
        (float, int),
    ),
    Option(
        "vedo",
        "cmap_n_colors",
        "Set the number of available colors",
        (int,),
    ),
    Option(
        "vedo",
        "scalarbar",
        "Scalarbar describing cmap. At least an empty dict or "
        "dict with following items are accepted: "
        "{title: str, pos: tuple, title_yoffset: int, font_size: int, "
        "nlabels: int, c: str, horizontal: bool, use_alpha: bool, "
        "label_format: str}. Setting bool will add a default scalarbar",
        (bool, dict),
    ),
    Option(
        "vedo",
        "scalarbar3d",
        "3D scalarbar describing cmap. At least an empty dict or "
        "dict with following items are accepted: "
        "{title: str, pos: tuple, size: list, title_font: str, "
        "title_xoffset: float, title_yoffset: float, title_size: float, "
        "title_rotation: float, nlabels: int, label_font: str, "
        "label_size: float, label_offset: float, label_rotation: int, "
        "label_format: str, draw_box: bool, above_text: str, below_text: str, "
        "nan_text: str, categories: list}",
        (bool, dict),
    ),
    Option(
        "vedo",
        "arrow_data",
        "Name of vertex_data to plot as arrow. Corresponding data should be "
        "at least 2D. If you want more control over arrows, consider creating "
        "edges using gus.create.edges.from_data().",
        (str,),
    ),
    Option(
        "vedo",
        "arrow_data_scale",
        "Scaling factor for arrow data.",
        (float, int),
    ),
    Option(
        "vedo",
        "arrow_data_color",
        "Color for arrow data. Can be either cmap name or color. For "
        "cmap, colors are based on the size of the arrows.",
        (str, tuple, list, int),
    ),
    Option(
        "vedo",
        "arrow_data_to_origin",
        "Points arrow data to geometric origin. By default, arrows point away "
        "from origin. "
        "When enabled, arrows are shifted backwards by their own magnitudes. "
        "Works in conjunction with arrow_data and arrow_data_scale options.",
        (bool,),
    ),
    Option(
        "vedo",
        "axes",
        "Configure a specific axes with options. Expect dict(), but setting "
        "True will set a default axes. For full options, see "
        "https://vedo.embl.es/autodocs/content/vedo/addons.html"
        "#vedo.addons.Axes",
        (bool, dict),
    ),
)


def make_valid_options(*options):
    """
    Forms valid options. Should run only once during module loading.

    Parameters
    ----------
    *options: Option

    Returns
    -------
    valid_options: dict()
    """
    valid_options = {}
    for opt in options:
        if isinstance(opt, Option):
            # copy option object to avoid overwriting defaults
            # only exception is if this option is a backend object
            # and wrapped by ModuleImportRaiser
            allowed_types = []
            for at in opt.allowed_types:
                if isinstance(at, ModuleImportRaiser):
                    continue
                allowed_types.append(at)
            opt.allowed_types = tuple(allowed_types)

            valid_options[opt.key] = deepcopy(opt)
        elif isinstance(opt, SetDefault):
            # overwrite default of existing option.
            if opt.key not in valid_options:
                raise KeyError("Given key is not in valid_option.")

            if type(opt.default) not in valid_options[opt.key].allowed_types:
                raise TypeError("Invalid default type")
            valid_options[opt.key].default = opt.default
        else:
            raise TypeError("Please use `Option` to define options.")

    return valid_options


class ShowOption(HelperBase):
    """
    Behaves similar to dict, but will only accept a set of options that's
    applicable to the helpee class. Intended use is to create a
    subclass that would define valid options for helpee.
    Options should be described by Option object.
    Helps all the way up to initializing backend showables up to their backend
    specific common routines. ShowOption and ShowManager in a sense.
    """

    __slots__ = ("_options",)

    _valid_options = {}

    _helps = None

    def __init__(self, helpee):
        """
        Parameters
        ----------
        helpee: object
        """
        self._helpee = helpee
        if not type(helpee).__qualname__.startswith(self._helps):
            raise TypeError(
                f"This show option is for {self._helps}. "
                f"Given helpee is {type(helpee)}."
            )
        self._options = {}

    def __repr__(self):
        """
        Friendly info prints for show option.

        Parameters
        ----------
        None

        Returns
        -------
        description: str
        """
        valid_and_current = []
        for vo in self._valid_options.values():
            valid = str(vo)
            current = ""
            if vo.key in self.keys():
                current = "current option: " + str(self[vo.key])
            valid_and_current.append(valid + current)

        header = [
            f"ShowOption for {self._helps}",
        ]
        return "\n".join(header + valid_and_current)

    def __setitem__(self, key, value):
        """
        Sets option after checking its validity.

        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        if key in self._valid_options:
            # valid type check
            if not isinstance(value, self._valid_options[key].allowed_types):
                raise TypeError(
                    f"{type(value)} is invalid value type for '{key}'. "
                    f"Details for '{key}':\n"
                    f"{self._valid_options[key]}"
                )

            # types are valid. let's add
            self._options[key] = value

        else:
            raise ValueError(f"{key} is an invalid option for {self._helps}.")

    def __getitem__(self, key):
        """
        operator[]

        Parameters
        ----------
        key: str or iterable

        Returns
        -------
        items: object or dict
        """
        if isinstance(key, str):
            return self._options[key]
        elif hasattr(key, "__iter__"):
            items = {}
            for k in key:
                if k in self._options:
                    items[k] = self._options[k]
            return items
        else:
            raise TypeError(f"Invalid key type for {type(self)}")

    def __contains__(self, key):
        """Returns if current option contains given key.

        Parameters
        ----------
        key: str

        Returns
        -------
        contains: bool
        """
        return key in self._options

    def __call__(self, **kwargs):
        """A short-cut to update(), but it returns helpee object.
        This is mainly to support one-line/inline visualizations.

        Parameters
        ----------
        kwargs: **kwargs

        Returns
        -------
        helpee: Any

        Example
        -------
        .. code-block:: python

            gus.show(
                mesh1.show_options(c="red"),
                mesh2.show_options(c="green"),
                mesh3.show_options(c="blue"),
            )
        """
        self.update(**kwargs)

        return self._helpee

    def get(self, key, default=None):
        """
        Gets value from key and default. Similar to dict.get(),
        but this is always safe, as it will always return None

        Parameters
        ----------
        key: stir
        default: object

        Returns
        -------
        values: object
        """
        if default is not None:
            return self._options.get(key, default)

        # overwrite default with valid option's
        default = getattr(self._valid_options.get(key, None), "default", None)

        return self._options.get(key, default)

    def update(self, **kwargs):
        """
        Calls __setitem__ iteratively for validity check.

        Parameters
        ----------
        **kwargs: kwargs

        Returns
        -------
        None
        """
        for k, v in kwargs.items():
            self.__setitem__(k, v)

    def valid_keys(self):
        """
        Returns valid keys.

        Parameters
        ----------
        None

        Returns
        -------
        valid_keys: dict_keys
        """
        return self._valid_options.keys()

    def keys(self):
        """
        Registered option keys.

        Parameters
        ----------
        None

        Returns
        -------
        keys: dict_keys
        """
        return self._options.keys()

    def values(self):
        """
        Registered option values.

        Parameters
        ----------
        None

        Returns
        -------
        keys: dict_values
        """
        return self._options.values()

    def items(self):
        """
        Registered option items.

        Parameters
        ----------
        None

        Returns
        -------
        items: dict_items
        """
        return self._options.items()

    def clear(self):
        """
        Clears all the options.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._options.clear()

    def pop(self, *args, **kwargs):
        """
        Calls pop() on current options

        Parameters
        ----------
        None

        Returns
        -------
        value: object
        """
        self._options.pop(*args, **kwargs)

    def copy_valid_options(self, copy_to, keys=None):
        """
        Copies valid option to other show_option. Simply iterates and tries.

        Parameters
        ----------
        copy_to: ShowOption
        keys: tuple or list
          Can specify keys

        Returns
        -------
        None
        """
        if not isinstance(copy_to, ShowOption):
            raise TypeError("copy_to should be a ShowOption")
        valid_keys = copy_to.valid_keys()

        items = self[keys].items() if keys is not None else self.items()

        for key, value in items:
            if key in valid_keys:
                copy_to[key] = deepcopy(value)  # is deepcopy necessary?

    def _initialize_showable(self):
        """
        Creates basic showable all the way up to backend common procedures.

        Parameters
        ----------
        None

        Returns
        -------
        showable: object
        """
        raise NotImplementedError("Derived class must implement this method")



from gustaf import (
    _version,
    create,
    edges,
    faces,
    helpers,
    io,
    settings,
    show,
    utils,
    vertices,
    volumes,
)
from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.vertices import Vertices
from gustaf.volumes import Volumes

__version__ = _version.version

__all__ = [
    "__version__",
    "settings",
    "vertices",
    "edges",
    "faces",
    "volumes",
    "show",
    "utils",
    "create",
    "io",
    "helpers",
    "Vertices",
    "Edges",
    "Faces",
    "Volumes",
]



from gustaf.utils import arr, connec, log, tictoc
from gustaf.utils.tictoc import Tic

# Alias
connectivity = connec

__all__ = [
    "arr",
    "connec",
    "connectivity",
    "log",
    "tictoc",
    "Tic",
]



"""gustaf/gustaf/io/mixd.py.

io functions for mixd.
"""

import os
import struct

import numpy as np

from gustaf import settings
from gustaf.faces import Faces
from gustaf.io.ioutils import abs_fname, check_and_makedirs
from gustaf.utils import log
from gustaf.utils.arr import close_rows
from gustaf.volumes import Volumes

_big_endian_int = ">i"
_big_endian_double = ">d"


def load(
    simplex=True,
    volume=False,
    fname=None,
    mxyz=None,
    mien=None,
    mrng=None,
):
    """mixd load. To avoid reading minf, all the crucial info can be given as
    params. Default input will try to import `mxyz`, `mien`, `mrng` from
    current location and assumes mesh is 2D triangle.

    Parameters
    -----------
    simplex: bool
      Default is True. Is it triangle based?
    volume: bool
      Default is False. Is it 3D?
    fname: str
      Default is None. Specify your mixd file names with ".xns" postfix.
      Ex) "gustaf.xns" will try to load "gustaf.mxyz", "gustaf.mien",
      and "gustaf.mrng"
    mxyz: str
      Default is None.
    mien: str
      Default is None.
    mrng: str
      Default is None. This is optional.

    Returns
    --------
    mesh: Faces or Volumes
    """
    # figure out input type
    specified_input = mxyz is not None  # bare minimum input
    fname_input = (fname is not None) and not specified_input
    default_input = not (fname_input or specified_input)

    if default_input:
        mxyz = "mxyz"
        mien = "mien"
        mrng = "mrng"

    elif fname_input:
        fname = abs_fname(fname)
        fbase, ext = os.path.splitext(fname)

        if os.path.basename(fbase) == "_":
            fbase = fbase[:-1]
        else:
            fbase += "."

        mxyz = fbase + "mxyz"
        mien = fbase + "mien"
        mrng = fbase + "mrng"

    # vertices
    vertices = np.fromfile(mxyz, dtype=_big_endian_double).astype(np.float64)

    # connec
    connec = None
    try:
        connec = (np.fromfile(mien, dtype=_big_endian_int) - 1).astype(
            np.int32
        )
    except BaseException:
        log.debug(f"mien file, `{mien}`, does not exist. Skipping.")

    # boundary conditions
    bcs = {}
    try:
        bcs_in = np.fromfile(mrng, dtype=_big_endian_int).astype(np.int32)
        uniq_bcs_in = np.unique(bcs_in)
        uniq_bcs_in = uniq_bcs_in[uniq_bcs_in > 0]  # keep only natural nums
        sub_elem_ids = np.arange(bcs_in.size)

        for ubci in uniq_bcs_in:
            bcs.update({str(ubci): sub_elem_ids[bcs_in == ubci]})

    except BaseException:
        log.debug(f"mrng file, `{mrng}`, does not exist. Skipping.")

    # reshape vertices
    vertices = vertices.reshape(-1, 3) if volume else vertices.reshape(-1, 2)

    # reshape connec
    if connec is not None:
        ncol = 3 if simplex and not volume else 4
        ncol = 8 if ncol == 4 and volume and not simplex else ncol

        connec = connec.reshape(-1, ncol)

        mesh = Volumes(vertices, connec) if volume else Faces(vertices, connec)

    # bc
    if len(bcs) != 0:
        mesh.BC = bcs

    return mesh


def _ravel(array):
    """
    Ravel if it is an array. Else, just return
    """
    if isinstance(array, np.ndarray):
        return array.ravel()
    return array


def _write_raveled_int(file, array):
    """
    Writes raveled array to the file as big endian int.

    Parameters
    ----------
    file: _io.TextIOWrapper
      Objects for calling `open(fname)`
    array: np.ndarray
    """
    for v in _ravel(array):
        file.write(struct.pack(_big_endian_int, v))


def _write_raveled_double(file, array):
    """
    Writes raveled array to the file as big endian double.

    Parameters
    ----------
    file: _io.TextIOWrapper
      Objects for calling `open(fname)`
    array: np.ndarray
    """
    for v in _ravel(array):
        file.write(struct.pack(_big_endian_double, v))


def export(
    fname,
    mesh,
    space_time=False,
    dual=False,
):
    """Export in mixd format. Supports triangle, quadrilateral, tetrahedron,
    and hexahedron semi-discrete and (flat) space-time mesh output.

    Parameters
    -----------
    mesh: Faces or Volumes
    fname: str
    space_time: bool
      Export Mesh as Space-Time Slab for discontinuous space-time
    dual: bool
      Includes dual-subelement information.

    Returns
    --------
    None
    """
    # did you give us an acceptable mesh?
    acceptable_shapes = ["tri", "quad", "tet", "hexa"]
    whatami = mesh.whatami

    if whatami not in acceptable_shapes:
        raise NotImplementedError(
            f"Sorry, we can't export {whatami}-shape in mixd format."
        )

    # prepare export location
    fname = abs_fname(fname)
    check_and_makedirs(fname)

    # basic infos
    dim = mesh.vertices.shape[1]

    # prep files
    fbase, ext = os.path.splitext(fname)

    if ext.startswith(".xns"):
        # frequently used case in practice. no base export
        if os.path.basename(fbase) == "_":
            fdir = os.path.dirname(fbase)
            vert_file = os.path.join(fdir, "mxyz")
            connec_file = os.path.join(fdir, "mien")
            bc_file = os.path.join(fdir, "mrng")
            info_file = os.path.join(fdir, "minf")
            dual_file = os.path.join(fdir, "dual")

        else:
            fbase += "."
            vert_file = fbase + "mxyz"
            connec_file = fbase + "mien"
            bc_file = fbase + "mrng"
            info_file = fbase + "minf"
            dual_file = fbase + "dual"

    else:
        raise NotImplementedError("`mixd` format only supports xns.")

    # write v
    with open(vert_file, "wb") as vf:
        _write_raveled_double(vf, mesh.vertices)

        # write it one more time for Discontinuous Prismic Space Time meshes
        # used to be called Flat Space Time mesh.
        if space_time:
            _write_raveled_double(vf, mesh.vertices)

    # write connec
    with open(connec_file, "wb") as cf:
        # let's not forget fortran numbering (+ 1)
        _write_raveled_int(cf, mesh.elements.ravel() + 1)

    # get boundaries of each element - subelement interface array
    sub_interface = make_mrng(mesh)

    # write bc first - after writing it, we can modify inplace for dual.
    with open(bc_file, "wb") as bf:
        _write_raveled_int(bf, sub_interface)

    # if dual is True, we fill dual infos.
    if dual:
        sub_elements = mesh.to_subelements(False)
        # this should be always dividable without remnants
        n_subelem_per_elem, rem = divmod(
            len(sub_elements.elements), len(mesh.elements)
        )
        if rem != 0:
            raise ValueError(
                "something went wrong with subelement creation."
                "Please report this issue, thank you!"
            )

        # get intersection - can use this info to determine duals
        _, _, _, intersections = close_rows(
            sub_elements.centers(), settings.TOLERANCE, True
        )

        # loop intersections and look for 2 intersections
        for i, intersection in enumerate(intersections):
            n_inter = len(intersection)

            # we modify interface only if there're 2 intersections.
            if n_inter == 2:
                # intersection is always sorted.
                # we don't want dual to point to itself
                dual_id = 0 if i != intersection[0] else 1

                # get element number and apply fortran's offset, 1
                sub_interface[i] = -int(
                    intersection[dual_id] // n_subelem_per_elem + 1
                )
                continue

            # intersection should be at most 2. Otherwise, it either means
            # that you have a bad mesh or to big tolerance
            if n_inter > 2:
                raise ValueError(
                    f"{i}-th subelement overlaps more than once. "
                    "Please check your elements or decrease "
                    "gustaf.settings.TOLERANCE."
                )

        # write dual
        with open(dual_file, "wb") as df:
            _write_raveled_int(df, sub_interface)

    # write info
    with open(info_file, "w") as infof:  # if and inf... just can't
        infof.write(f"# dim: {dim}\n")
        infof.write(f"# mesh type: {whatami}\n\n")

        st_factor = 2 if space_time else 1
        infof.write(f"nn {int(mesh.vertices.shape[0] * st_factor)}\n")
        infof.write(f"ne {int(mesh.elements.shape[0])}\n")
        infof.write(f"nsd {dim}\n")
        infof.write(f"nen {int(mesh.elements.shape[1] * st_factor)}\n")

        if space_time:
            infof.write("space-time on\n\n\n")
        else:
            infof.write("semi-discrete on\n\n\n")

        # BC guide
        infof.write("# boundary name : referenced number.\n")
        for i, bname in enumerate(mesh.BC.keys()):
            infof.write(f"# {bname} : {i + 1}\n")

        # signature
        infof.write("\n\n\n# MIXD generated using `gustaf`.\n")


def make_mrng(mesh):
    """
    Builds and return mrng array based on `mesh.BC`
    Supports `Faces` and `Volumes`.

    Parameters
    -----------
    mesh: Faces or Volumes
      Number of participating elements

    Returns
    --------
    boundaries : ndarray
      The mrng-array.
    """

    # determine number of sub elements
    whatami = mesh.whatami
    nbelem = 3

    if whatami.startswith("quad") or whatami.startswith("tet"):
        nbelem += 1
    elif whatami.startswith("hexa"):
        nbelem += 3

    # init boundaries with 0, as boundaries are marked with positive numbers
    # staring from 1 and dual infos with negative values, starting with -1
    boundaries = np.empty(mesh.elements.shape[0] * nbelem, dtype=int)
    boundaries[:] = 0

    for i, belem_ids in enumerate(mesh.BC.values()):
        boundaries[belem_ids] = i + 1  # bid starts at 1

    return boundaries



"""gustaf/gustaf/helpers/_base.py

Base class for helper
"""

from weakref import ref

from gustaf._base import GustafBase


class HelperBase(GustafBase):
    """
    Minimal base layer for helper classes to avoid cyclic referencing.
    Instead of saving a pure reference to helpee object, this will create a
    weakref instead. This class defines `_helpee` setters and getters, so that
    existing code can remain untouched.
    """

    __slots__ = ("_helpee_weak_ref",)

    @property
    def _helpee(self):
        """Returns dereferenced weak ref. Setter will create and save weakref
        of a helpee."""
        return self._helpee_weak_ref()

    @_helpee.setter
    def _helpee(self, helpee):
        self._helpee_weak_ref = ref(helpee)



"""Import a meshio based mesh,

Export can only happen after it is possible to save and define boundaries in
gustaf.
"""

import pathlib

import numpy as np

from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.helpers.raise_if import ModuleImportRaiser
from gustaf.utils import log
from gustaf.vertices import Vertices
from gustaf.volumes import Volumes

try:
    import meshio
except ModuleNotFoundError as err:
    meshio = ModuleImportRaiser("meshio", err)

_meshio2gus = {
    "hexahedron": Volumes,
    "tetra": Volumes,
    "quad": Faces,
    "triangle": Faces,
    "line": Edges,
    "vertex": Vertices,
}


def load(fname):
    """Load mesh in meshio format. Loads vertices and their connectivity.
    Currently cannot process boundary.

    Note
    -----
    This is more or less a direct copy from the original gustav implementation.
    A lot of the meshio information are lost. When boundaries and multi-patch
    definitions are added this needs to be revisited and extended.

    Parameters
    ------------
    fname: str | pathlib.Path

    Returns
    --------
    MESH_TYPES | List[MESH_TYPES]
    """
    # fname sanity check
    fname = pathlib.Path(fname)
    if not (fname.exists() and fname.is_file()):
        raise ValueError(
            "The given file does not point to file. The given path is: "
            f"{fname.resolve()}"
        )

    # load
    meshio_mesh: meshio.Mesh = meshio.read(fname)

    # first get vertices
    vertices = meshio_mesh.points

    # early exit if cells doesn't exist
    if len(meshio_mesh.cells_dict) == 0:
        return Vertices(vertices)

    meshes = []
    for element_type, elements in meshio_mesh.cells_dict.items():
        # skip unsupported
        if element_type not in _meshio2gus:
            log.warning(
                f"`{element_type}`-elements are not supported in gustaf"
            )
            continue
        if element_type.startswith("vertex"):
            meshes.append(Vertices(vertices[elements.ravel()]))
        else:
            meshes.append(
                _meshio2gus[element_type](vertices, elements=elements)
            )

    return meshes[0] if len(meshes) == 1 else meshes


def export(fname, mesh, submeshes=None, **kwargs):
    """Export mesh elements and vertex data into meshio and use its write
    function. The definition of submeshes with identical vertex coordinates
    is possible. In that case vertex numbering and data from the main mesh
    are used. For more export options, refer to meshio's documentation
    https://github.com/nschloe/meshio .

    .. code-block:: python

        import gustaf

        # define coordinates
        v = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [1.0, 0.0, 1.0],
                [0.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
            ]
        )
        # define triangle connectivity
        tf = np.array(
            [
                [1, 0, 2],
                [0, 1, 5],
                [3, 2, 6],
                [2, 0, 4],
                [4, 5, 7],
                [2, 3, 1],
                [7, 5, 1],
                [6, 7, 3],
                [4, 6, 2],
                [7, 6, 4],
            ]
        )
        # init tri faces
        mesh = gus.Faces(
            vertices=v,
            faces=tf,
        )
        gustaf.io.meshio.export(mesh, "tri-mesh.stl")

    Parameters
    ------------
    fname: Union[str, pathlib.Path]
      File to save the mesh in.
    mesh: Edges, Faces or Volumes
      Input mesh
    submeshes: Iterable
      Submeshes where the vertices are identical to the main mesh. The element
      type can be identical to mesh.elements or lower-dimensional (e.g.
      boundary elements).
    **kwargs : Any
      Any additional argument will be passed to the respective meshio `write`
      function. See meshio docs for more information

    Raises
    -------
    NotImplementedError:
       For mesh types that are not implemented.
    ValueError:
       Raises a value error, if the vertices indexed in a subset are not
       present in the main mesh.
    """

    # Mapping between meshio cell types and gustaf cell types
    meshio_dict = {
        "edges": "line",
        "tri": "triangle",
        "quad": "quad",
        "tet": "tetra",
        "hexa": "hexahedron",
    }

    cells = []
    # Merge main mesh and submeshes in one list
    meshes = mesh if isinstance(mesh, list) else [mesh]

    if submeshes is not None:
        meshes.extend(submeshes)

    # Iterate the meshes and extract the element information in meshio format
    for m in meshes:
        whatami = m.whatami
        if whatami not in meshio_dict:
            raise NotImplementedError(
                f"{whatami}-type meshes not supported (yet)."
            )
        elif np.any(m.elements > len(m.vertices) - 1):
            raise ValueError("Invalid vertex IDs in submesh connectivity.")
        else:
            cells.append((meshio_dict[whatami], m.elements))

    # Export data to meshio and write file
    meshio.Mesh(
        points=mesh.vertices,
        cells=cells,
        point_data=mesh.vertex_data._saved,
    ).write(fname, **kwargs)



"""gustaf/create/edges.py
Routines to create edges.
"""

import numpy as np

from gustaf import utils
from gustaf.edges import Edges


def from_vertices(vertices, closed=False, continuous=True):
    """
    Creates Edges with given vertices. If close==True,
    last vertices will be connected to the first one.
    If continuous==False, it assumes that every two vertices form
    an independent line.

    Parameters
    ----------
    vertices: (n, d) np.ndarray or Vertices
    close: bool
    continuous: bool

    Returns
    -------
    edges: Edges
    """
    if hasattr(vertices, "vertices"):  # noqa SIM108
        v = vertices.vertices
    else:
        v = vertices

    edges = Edges(v, utils.connec.range_to_edges(len(v), closed, continuous))

    return edges


def from_data(gus_obj, data, scale=None, data_norm=None):
    """
    Creates edges from gustaf object with vertices.
    Data can be either multi-dim array-like data or a str describing a name
    of vertex_data that belongs to given gustaf object.
    len(gus_obj.vertices) number of edges will be created, where origin and
    end of each edge is created using the following scheme:
    [[vertices[0], vertices[0] + (array_data[0] * scale)], ...].
    By default, scaling value will be
    max([1, (aabb_diagonal_norm * 0.1 / max_data_norm)]).
    If there's dimension mismatch between vertices and the data, will append
    zero paddings!

    Parameters
    ----------
    gus_obj: Vertices
      gus.Vertices or its derived classes
    data: str or (n_vertices, d) array-like
     If str, will be considered as data and search for saved vertex_data.
    scale: float
      Absolute value.
    data_norm: float or array-like
      If float, will be considered as max_norm of the data. Else, searches for
      max value. Doesn't enforce len to match.

    Returns
    -------
    data_arrow: Edges
    """
    if not isinstance(gus_obj, Edges.__boundary_class__):
        raise TypeError(
            "Invalid input. Expecting gus.Vertices or its subclasses"
        )

    # get origin and increment (= array_data * scale)
    origin = gus_obj.const_vertices
    if isinstance(data, str):
        # will raise if data doesn't exist
        increment = gus_obj.vertex_data[data]
    elif isinstance(data, (tuple, list, np.ndarray)):
        increment = np.asanyarray(data)
        if len(increment) != len(origin):
            raise ValueError(
                f"Data length mismatch: expected ({len(origin)}) / "
                "given ({len(increment)})"
            )
    else:
        raise TypeError(f"Couldn't process {type(data)}-data as input.")

    # data dim check
    inc_dim = increment.shape[1]
    origin_dim = origin.shape[1]
    if inc_dim == 1:
        # if data is scalar, it should be addable to any vertices
        utils.log.debug(
            f"gus.create.edges.from_data() - requested data ({data}) is",
            "scalar. Edge orientation will be [1,1,..,1].",
        )
    elif inc_dim != origin_dim:
        # match data and vertex dim.
        utils.log.debug(
            "gus.create.edges.from_data() - dimension mismatch between",
            f"vertices ({origin_dim}) and data ({inc_dim}). Matching",
            "dimension by appending zeros.",
        )
        dim_diff = int(inc_dim - origin_dim)
        zero_pad = np.zeros((len(origin), abs(dim_diff)))
        if dim_diff > 0:
            origin = np.hstack((origin, zero_pad))
        else:
            increment = np.hstack((increment, zero_pad))

    # apply default scale
    if scale is None:
        if isinstance(data, str):
            norm = gus_obj.vertex_data.as_scalar(data, None, True)
        elif data_norm is None:
            norm = np.linalg.norm(increment, axis=1)
        else:
            norm = np.asanyarray(data_norm)

        max_data_norm = norm.max()
        aabb_diagonal_norm = gus_obj.bounds_diagonal_norm()
        scale = min((1.0, aabb_diagonal_norm * (0.1) / max_data_norm))
        utils.log.debug(
            f"creating edges from data with scaling factor ({scale})"
        )

    # by here, this should be good to go
    vs = np.hstack((origin, origin + (increment * scale))).reshape(
        -1, origin.shape[1]
    )

    return Edges(vs, utils.connec.range_to_edges(len(vs), continuous=False))



"""gustaf/gustaf/helpers/notebook.py.

Enables the plotting in ipynb with k3d.

There are no import guards since they are in the place where this module is
imported. This should be enough since I do not think that this module
will/should be used outside this place.
"""

import importlib

import numpy as np
import vedo
from IPython.display import display
from ipywidgets import GridspecLayout

from gustaf._base import GustafBase


def get_shape(N, x, y):
    """Taken verbatim from vedo plotter:show function.

    Args:
        N (_type_): _description_
        x (_type_): _description_
        y (_type_): _description_

    Returns:
        _type_: _description_
    """
    nx = int(np.sqrt(int(N * y / x) + 1))
    ny = int(np.sqrt(int(N * x / y) + 1))
    lm = [
        (nx, ny),
        (nx, ny + 1),
        (nx - 1, ny),
        (nx + 1, ny),
        (nx, ny - 1),
        (nx - 1, ny + 1),
        (nx + 1, ny - 1),
        (nx + 1, ny + 1),
        (nx - 1, ny - 1),
    ]
    ind, minl = 0, 1000
    for i, m in enumerate(lm):
        l_something = m[0] * m[1]
        if N <= l_something < minl:
            ind = i
            minl = l_something
    return lm[ind]


class K3DPlotterN(GridspecLayout, GustafBase):
    """Helper to plot in notebooks with k3d.

    Sets up K3D plotter especially if multiple plots are to be shown in the
    notebook.
    """

    def __init__(self, N, size, background=0xFFFFFF):
        """Setups the plotter with potentially multiple plots.

        Parameters
        ----------
        N: int
            Number of plots to be shown.
        size: (str, list)
            See vedo.Plotter for details.
        background: ((color, str), optional)
          See vedo.Plotter for details. Defaults to 0xFFFFFF.
        """
        if importlib.util.find_spec("k3d") is None:
            raise ImportError(
                "k3d is not installed. Please install it with `pip "
                "install k3d`."
            )
        self.N = N
        self.x, self.y = get_shape(N, *(2160, 1440))
        self.shape = (self.x, self.y)
        super().__init__(self.x, self.y)
        self.vedo_plotters = []
        for _ in range(N):
            self.vedo_plotters.append(
                vedo.Plotter(
                    size=size,
                    bg=background,
                )
            )

    def _at_get_location(self, N):
        """Gets the grid coordinate of the wanted renderer.

        Parameters
        ----------
        N: int
            Render id.

        Returns:
            (int, int): Grid Coordinate of the wanted renderer.
        """
        if (self.x * self.y) < N:
            return (self.x - 1, self.y - 1)
        return (N // (self.y), N % self.y)

    def show(
        self, list_of_showables, at, interactive, camera, axes, *args, **kwargs
    ):
        """Add the showables to the renderer at the given location.

        Parameters
        -----------
        list_of_showables: Any
        at: int
            Render id.
        interactive: bool
            See vedo.Plotter.show for details.
        camera: Any
            See vedo.Plotter.show for details.
        axes: bool
            Add axes to the plot. Will also cast int to bool.
        """
        if len(args) != 0 or len(kwargs) != 0:
            self._logd(f"*args ({args}) and **kwargs ({kwargs}) ignored")

        # this converts vedo plotter to k3d plot.
        # after this, vedo plotter has no relevance to what you see
        self[self._at_get_location(at)] = self.vedo_plotters[at].show(
            list_of_showables,
            interactive=interactive,
            camera=camera,
            axes=axes,
        )

    def display(self, close=True):
        """Display the plotter.

        This is needed in case the plotter is the last thing in a cell. In that
        case the IPython will try to call this function to display this.
        """
        display(self)

        # we add option to close here, as we set default_autoclose = False
        if close:
            self.clear()
            self.close()

    def clear(self, *args, **kwargs):
        """Clear the plotters."""
        for v_plotter in self.vedo_plotters:
            v_plotter.clear(*args, deep=True, **kwargs)

    def close(self):
        """Closes all vedo.Plotters"""
        for v_plotter in self.vedo_plotters:
            v_plotter.close()



import numpy as np
import pytest

# frequently used fixtures
all_grids = (
    "vertices",
    "edges",
    "faces_tri",
    "faces_quad",
    "volumes_tet",
    "volumes_hexa",
)


@pytest.mark.parametrize("grid", all_grids)
def test_concat(grid, request):
    """Test concat. uses merge_vertices to double check."""
    grid = request.getfixturevalue(grid)

    n_grids = 5

    n_vertices = len(grid.vertices)

    if grid.kind != "vertex":
        n_elements = len(grid.elements)

    grids = [grid for _ in range(n_grids)]

    concated = type(grid).concat(*grids)

    # check concatenated vertices
    assert np.allclose(np.tile(grid.vertices, (n_grids, 1)), concated.vertices)

    # check number of concatenated elements
    if grid.kind != "vertex":
        assert int(n_elements * n_grids) == len(concated.elements)
        assert int(n_vertices * n_grids - 1) == concated.elements.max()

    # merge vertices, then this should only have n_vertices left
    concated.merge_vertices()
    assert np.allclose(grid.vertices, concated.vertices)

    # as well as n_elements * n_grids, but elements just repeats n_grids times
    if grid.kind != "vertex":
        assert (
            np.tile(grid.elements, (n_grids, 1)) - concated.elements
        ).sum() == 0



"""gustaf/gustaf/settings.py.

Global variables/constants that's used throughout `gustaf`.
"""

TOLERANCE = 1e-10

FLOAT_DTYPE = "float64"
INT_DTYPE = "int32"

# OPTIONS are <"vedo" | "trimesh" | "matplotlib">
VISUALIZATION_BACKEND = "vedo"

VEDO_DEFAULT_OPTIONS = {
    "vertex": {},
    "edges": {},
    "faces": {},
    "volumes": {},
}

NTHREADS = 1



import os
import tempfile

import numpy as np
import pytest

import gustaf as gus

all_grids = (
    ("faces_tri_2d", "mfem_triangles_2d.mesh"),
    ("faces_quad_2d", "mfem_quadrilaterals_2d.mesh"),
    ("volumes_hexa", "mfem_hexahedra_3d.mesh"),
    ("volumes_tet", "mfem_tetrahedra_3d.mesh"),
    ("volumes_hexa_222", "mfem_hexahedra_3d_222.mesh"),
)


@pytest.mark.parametrize(("grid", "ground_truth_filename"), all_grids)
def test_mfem_export(
    to_tmpf,
    are_stripped_lines_same,
    grid,
    ground_truth_filename,
    request,
):
    mesh = request.getfixturevalue(grid)

    verts = mesh.vertices

    if mesh.whatami in ("tri", "quad"):
        lines = mesh.to_edges(unique=False)
        BC = {1: [], 2: [], 3: []}
        for i in lines.single_edges():
            # mark boundaries at x = 0 with 1
            if np.max(verts[lines.const_edges[i], 0]) < 0.1:
                BC[1].append(i)
            # mark boundaries at x = 1 with 2
            elif np.min(verts[lines.const_edges[i], 0]) > 0.9:
                BC[2].append(i)
            # mark rest of the boundaries with 3
            else:
                BC[3].append(i)
    elif mesh.whatami in ("hexa", "tet"):
        faces = mesh.to_faces(unique=False)

        BC = {1: [], 2: [], 3: []}
        # single faces only produces exterior/boundary faces
        for i in faces.single_faces():
            # mark boundaries at x = 0 with 1
            if np.max(verts[faces.const_faces[i], 0]) < 0.1:
                BC[1].append(i)
            # mark boundaries at x = 1 with 2
            elif np.min(verts[faces.const_faces[i], 0]) > 0.9:
                BC[2].append(i)
            # mark rest of the boundaries with 3 if they do not contain
            # interior vertices
            else:
                BC[3].append(i)

    mesh.BC = BC

    # Test output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        gus.io.mfem.export(tmpf, mesh)

        with (
            open(tmpf) as tmp_read,
            open(
                os.path.join(
                    os.path.dirname(__file__),
                    f"data/{ground_truth_filename}",
                )
            ) as base_file,
        ):
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), ignore_order=True
            )



"""gustaf/gustaf/utils/log.py.

Thin logging wrapper.
"""

import logging
from functools import partial, partialmethod


def configure(debug=False, logfile=None):
    """Logging configurator.

    Parameters
    -----------
    debug: bool
    logfile: str

    Returns
    --------
    None
    """
    # logger
    logger = logging.getLogger("gustaf")

    # level
    level = logging.DEBUG if debug else logging.INFO
    logger.setLevel(level)

    # format
    formatter = logging.Formatter(fmt="%(name)s [%(levelname)s] %(message)s")

    # apply format using stream handler
    # let's use only one stream handler so that calling configure multiple
    # times won't duplicate printing.
    new_handlers = []
    for _i, h in enumerate(logger.handlers):
        # we skip all the stream handler.
        if isinstance(h, logging.StreamHandler):
            continue

        # blindly keep other ones
        else:
            new_handlers.append(h)

    # add new stream handler
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter)
    new_handlers.append(stream_handler)

    logger.handlers = new_handlers

    # output logs
    if logfile is not None:
        file_logger_handler = logging.FileHandler(logfile)
        logger.addHandler(file_logger_handler)


def debug(*log):
    """Debug logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logger = logging.getLogger("gustaf")
    logger.debug(" ".join(map(str, log)))


def info(*log):
    """Info logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logger = logging.getLogger("gustaf")
    logger.info(" ".join(map(str, log)))


def warning(*log):
    """warning logger.

    Parameters
    -----------
    *log: Tuple[str]

    Returns
    --------
    None
    """
    logger = logging.getLogger("gustaf")
    logger.warning(" ".join(map(str, log)))


def prepended_log(message, log_func, as_method=True):
    """
    Prepend message before a logging function.

    Parameters
    ----------
    message: str
    log_func: function
      one of the following - {info, debug, warning}
    as_method: bool
      If True, uses partialmethod, else partial.
      Default is True.

    Returns
    -------
    prepended: function
    """
    p = partialmethod if as_method else partial
    return p(log_func, message)



"""gustaf/gustaf/utils/connec.py.

Useful functions for connectivity operation. Ranging from edges to
volumes. Named connec because connectivity is too long. Would have been
even cooler, if it was palindrome.
"""

import collections

import numpy as np

try:
    import napf
except ImportError:
    from gustaf.helpers.raise_if import ModuleImportRaiser

    napf = ModuleImportRaiser("napf")

from gustaf import helpers, settings
from gustaf.utils import arr


def tet_to_tri(volumes):
    """Computes tri faces based on following index scheme.

    ``Tetrahedron``

    .. code-block:: text

        Ref: (node_ind), face_ind

                      (0)
                     _/|
                   _/ 1|
            (1)  _/____| (3)
               _/|    /|
             _/ 0| 2/ 3|
            /____|/____|
          (0)   (2)    (0)

        face_ind | node_ind
        ---------|----------
        0        | 0 2 1
        1        | 1 3 0
        2        | 2 3 1
        3        | 3 2 0

    Parameters
    -----------
    volumes: (n, 4) np.ndarray

    Returns
    --------
    faces: (n * 4, 3) np.ndarray
    """
    volumes = np.asanyarray(volumes, settings.INT_DTYPE)

    if volumes.ndim != 2 or volumes.shape[1] != 4:
        raise ValueError("Given volumes are not `tet` volumes")

    fpe = 4  # faces per element
    faces = (
        np.ones(((volumes.shape[0] * fpe), 3), dtype=settings.INT_DTYPE) * -1
    )  # -1 for safety check

    faces[:, 0] = volumes.ravel()
    faces[::fpe, [1, 2]] = volumes[:, [2, 1]]
    faces[1::fpe, [1, 2]] = volumes[:, [3, 0]]
    faces[2::fpe, [1, 2]] = volumes[:, [3, 1]]
    faces[3::fpe, [1, 2]] = volumes[:, [2, 0]]

    if (faces == -1).any():
        raise ValueError("Something went wrong while computing faces")

    return faces


def hexa_to_quad(volumes):
    """Computes quad faces based on following index scheme.

    ``Hexahedron``

    .. code-block:: text


                (6)    (7)
                *------*
                |      |
         (6) (2)| 3    |(3)   (7)    (6)
         *------*------*------*------*
         |      |      |      |      |
         | 2    | 0    | 4    | 5    |
         *------*------*------*------*
         (5) (1)|      |(0)   (4)    (5)
                | 1    |
                *------*
                (5)    (4)

        face_ind | node_ind
        ---------|----------
        0        | 1 0 3 2
        1        | 0 1 5 4
        2        | 1 2 6 5
        3        | 2 3 7 6
        4        | 3 0 4 7
        5        | 4 5 6 7

    Parameters
    -----------
    volumes: (n, 8) np.ndarray

    Returns
    --------
    faces: (n * 8, 4) np.ndarray
    """
    volumes = np.asanyarray(volumes, settings.INT_DTYPE)

    if volumes.ndim != 2 or volumes.shape[1] != 8:
        raise ValueError("Given volumes are not `hexa` volumes")

    fpe = 6  # faces per element
    faces = np.empty(((volumes.shape[0] * fpe), 4), dtype=settings.INT_DTYPE)

    faces[::fpe] = volumes[:, [1, 0, 3, 2]]
    faces[1::fpe] = volumes[:, [0, 1, 5, 4]]
    faces[2::fpe] = volumes[:, [1, 2, 6, 5]]
    faces[3::fpe] = volumes[:, [2, 3, 7, 6]]
    faces[4::fpe] = volumes[:, [3, 0, 4, 7]]
    faces[5::fpe] = volumes[:, [4, 5, 6, 7]]

    return faces


def volumes_to_faces(volumes):
    """Guidance function for `tet_to_tri` and `hexa_to_quad`.

    Parameters
    -----------
    volumes: (n, 4) or (m, 8) np.ndarray

    Returns
    --------
    faces: (n*4, 3) or (m*6, 4) np.ndarray
    """
    volumes = np.asanyarray(volumes, settings.INT_DTYPE)
    if volumes.shape[1] == 4:
        return tet_to_tri(volumes)

    elif volumes.shape[1] == 8:
        return hexa_to_quad(volumes)


def faces_to_edges(faces):
    """Compute edges based on following edge scheme.

    .. code-block:: text

        Ref: (node_ind), edge_ind

            (0)
             /\\
          0 /  \\2
           /____\\
        (1)  1   (2)

              2
        (3)*-----*(2)
           |     |
         3 |     | 1
        (0)*-----*(1)
              0

    Note: if `edges` index matter for tets, reorder it!

    Parameters
    -----------
    faces: (n, 3) or (n, 4) np.ndarray

    Returns
    --------
    edges: (n * 3, 2) or (n * 4, 2) np.ndarray
    """
    if faces.ndim != 2:
        raise ValueError(
            "Given input has wrong dimension. "
            "The input array for a faces has to be dim of 2"
        )

    num_faces = faces.shape[0]
    vertices_per_face = faces.shape[1]

    num_edges = int(num_faces * vertices_per_face)
    edges = np.empty((num_edges, 2), dtype=settings.INT_DTYPE)

    edges[:, 0] = faces.ravel()

    for i in range(vertices_per_face):
        # v_ind : corresponding vertex index for i value
        v_ind = 0 if i == int(vertices_per_face - 1) else i + 1

        edges[i::vertices_per_face, 1] = faces[:, v_ind]

    return edges


def range_to_edges(range_, closed=False, continuous=True):
    """Given range, for example (a, b), returns an edge sequence that
    sequentially connects indices. If int is given as range, it is considered
    as (0, value). Used to be called "closed/open_loop_index_train".

    Parameters
    -----------
    range_: list, tuple, or int
    closed: bool
    continuous: bool

    Returns
    --------
    edges: (n, 2) np.ndarray
    """
    if isinstance(range_, int):
        indices = np.arange(range_, dtype=settings.INT_DTYPE)
    elif isinstance(range_, (list, tuple)):
        # pass range_ as is and check for valid output
        indices = np.arange(*range_, dtype=settings.INT_DTYPE)
        if len(indices) < 2:
            raise ValueError(
                f"{range_} is invalid range input. "
                "It must result in minimum of size=2 array."
            )

    # closed is ignored
    if not continuous:
        if indices.size % 2 == 1:
            raise ValueError(
                "Ranges should result in even number of indices for "
                "continuous edges."
            )
        return indices.reshape(-1, 2)

    return sequence_to_edges(indices, closed)


def sequence_to_edges(seq, closed=False):
    """Given a sequence of int, "connect" to turn them into edges.

    Parameters
    -----------
    seq: (n,) array-like
    closed: bool

    Returns
    --------
    edges: (m, 2) np.ndarray
    """
    edges = np.repeat(seq, 2)

    if closed:
        first = int(edges[0])  # this is redundant copy to ensure detaching
        # equivalent to np.roll(edges, -1) without copy
        # safe since numpy is single thread
        edges[:-1] = edges[1:]
        edges[-1] = first
    else:
        # reselect only part of the array instead of copy
        # this should only update array interface protocol
        edges = edges[1:-1]

    return edges.reshape(-1, 2)


def make_quad_faces(resolutions):
    """Given number of nodes per each dimension, returns connectivity
    information of a structured mesh. Counter clock wise connectivity.

    .. code-block:: text

        (3)*------*(2)
           |      |
           |      |
        (0)*------*(1)

    Parameters
    ----------
    resolutions: list

    Returns
    -------
    faces: (n, 4) np.ndarray
    """
    nnpd = np.asarray(resolutions)  # number of nodes per dimension
    if any(nnpd < 1):
        raise ValueError(f"The number of nodes per dimension is wrong: {nnpd}")

    total_nodes = np.prod(nnpd)
    total_faces = (nnpd[0] - 1) * (nnpd[1] - 1)
    try:
        node_indices = np.arange(
            total_nodes, dtype=settings.INT_DTYPE
        ).reshape(nnpd[1], nnpd[0])
    except ValueError as e:
        raise ValueError(f"Problem with generating node indices. {e}")

    faces = np.empty((total_faces, 4), dtype=settings.INT_DTYPE)

    faces[:, 0] = node_indices[: (nnpd[1] - 1), : (nnpd[0] - 1)].ravel()
    faces[:, 1] = node_indices[: (nnpd[1] - 1), 1 : nnpd[0]].ravel()
    faces[:, 2] = node_indices[1 : nnpd[1], 1 : nnpd[0]].ravel()
    faces[:, 3] = node_indices[1 : nnpd[1], : (nnpd[0] - 1)].ravel()

    return faces


def make_hexa_volumes(resolutions):
    """Given number of nodes per each dimension, returns connectivity
    information of structured hexahedron elements. Counter clock wise
    connectivity.

    .. code-block:: text

         (7)*-------*(6)
           /|      /|
          / | (5) / |
      (4)*-------*  |
         |  *----|--*(2)
         | /(3)  | /
         |/      |/
      (0)*-------*(1)

    Parameters
    -----------
    resolutions: list

    Returns
    --------
    elements: (n, 8) np.ndarray
    """
    nnpd = np.asarray(resolutions)  # number of nodes per dimension
    if any(nnpd < 1):
        raise ValueError(f"The number of nodes per dimension is wrong: {nnpd}")

    total_nodes = np.prod(nnpd)
    total_volumes = np.prod(nnpd - 1)
    node_indices = np.arange(total_nodes, dtype=settings.INT_DTYPE).reshape(
        nnpd[::-1]
    )

    volumes = np.empty((total_volumes, 8), dtype=settings.INT_DTYPE)

    volumes[:, 0] = node_indices[
        : (nnpd[2] - 1), : (nnpd[1] - 1), : (nnpd[0] - 1)
    ].ravel()
    volumes[:, 1] = node_indices[
        : (nnpd[2] - 1), : (nnpd[1] - 1), 1 : nnpd[0]
    ].ravel()
    volumes[:, 2] = node_indices[
        : (nnpd[2] - 1), 1 : nnpd[1], 1 : nnpd[0]
    ].ravel()
    volumes[:, 3] = node_indices[
        : (nnpd[2] - 1), 1 : nnpd[1], : (nnpd[0] - 1)
    ].ravel()
    volumes[:, 4] = node_indices[
        1 : nnpd[2], : (nnpd[1] - 1), : (nnpd[0] - 1)
    ].ravel()
    volumes[:, 5] = node_indices[
        1 : nnpd[2], : (nnpd[1] - 1), 1 : nnpd[0]
    ].ravel()
    volumes[:, 6] = node_indices[1 : nnpd[2], 1 : nnpd[1], 1 : nnpd[0]].ravel()
    volumes[:, 7] = node_indices[
        1 : nnpd[2], 1 : nnpd[1], : (nnpd[0] - 1)
    ].ravel()

    return volumes


def subdivide_edges(edges):
    """Subdivide edges. We assume that mid point is newly added points.

    ``Subdivided Edges``

    .. code-block:: text

        Edges (Lines)

        Ref: (node_ids), edge_ids

        (0)      (2)      (1)
         *--------*--------*

        edge_ids | node_ids
        ---------|----------
        0        | 0 2
        1        | 2 1

    Parameters
    -----------
    edges: (n, 2) np.ndarray

    Returns
    --------
    subdivided_edges: (n * 2, 2) np.ndarray
    """
    if edges.ndim != 2 or edges.shape[1] != 2:
        raise ValueError("Invalid edges shape!")

    raise NotImplementedError


def subdivide_tri(mesh, return_dict=False):
    """Subdivide triangles. Each triangle is divided into 4 meshes.

    ``Subdivided Faces``

    .. code-block:: text

        Triangles

        Ref: (node_ind), face_ind

                 (0)
                     _/|
                   _/ 0|
            (3)  _/____|(5)
               _/|    /|
             _/ 1| 3/ 2|
            /____|/____|
          (1)   (4)    (2)

        face_ind | node_ind
        ---------|----------
        0        | 0 3 5
        1        | 1 4 3
        2        | 2 5 4
        3        | 3 4 5

    Parameters
    -----------
    mesh: Mesh
    return_dict: bool

    Returns
    --------
    new_vertices: (n, d) np.ndarray
    subdivided_faces: (m, 3) np.ndarray
    mesh_dict: dict
      iff `return_dict=True`,
      returns dict(vertices=new_vertices, faces=subdivided_faces).
    """
    # This will only pass if the mesh is triangle mesh.
    if mesh.faces.shape[1] != 3:
        raise ValueError("Invalid faces shape!")

    # Form new vertices
    edge_mid_v = mesh.vertices[mesh.unique_edges().values].mean(axis=1)
    new_vertices = np.vstack((mesh.vertices, edge_mid_v))

    subdivided_faces = np.empty(
        (mesh.faces.shape[0] * 4, mesh.faces.shape[1]),
        dtype=settings.INT_DTYPE,
    )

    mask = np.ones(subdivided_faces.shape[0], dtype=bool)
    mask[3::4] = False

    # 0th column minus (every 4th row, starting from 3rd row)
    subdivided_faces[mask, 0] = mesh.faces.ravel()

    # Form ids for new vertices
    new_vertices_ids = mesh.unique_edges().inverse + int(mesh.faces.max() + 1)
    # 1st & 2nd columns
    subdivided_faces[mask, 1] = new_vertices_ids
    subdivided_faces[mask, 2] = new_vertices_ids.reshape(-1, 3)[
        :, [2, 0, 1]
    ].ravel()

    # Every 4th row, starting from 3rd row
    subdivided_faces[~mask, :] = new_vertices_ids.reshape(-1, 3)

    if return_dict:
        return {
            "vertices": new_vertices,
            "faces": subdivided_faces,
        }

    else:
        return new_vertices, subdivided_faces


def subdivide_quad(
    mesh,
    return_dict=False,
):
    """Subdivide quads.

    Parameters
    -----------
    mesh: Mesh
    return_dict: bool

    Returns
    --------
    new_vertices: (n, d) np.ndarray
    subdivided_faces: (m, 4) np.ndarray
    mesh_dict: dict
      iff `return_dict=True`,
      returns dict(vertices=new_vertices, faces=subdivided_faces).
    """
    # This will only pass if the mesh is quad mesh.
    if mesh.faces.shape[1] != 4:
        raise ValueError("Invalid faces shape!")

    # Form new vertices
    edge_mid_v = mesh.vertices[mesh.unique_edges().values].mean(axis=1)
    face_centers = mesh.centers()
    new_vertices = np.vstack(
        (
            mesh.vertices,
            edge_mid_v,
            face_centers,
        )
    )

    subdivided_faces = np.empty(
        (mesh.faces.shape[0] * 4, mesh.faces.shape[1]),
        dtype=settings.INT_DTYPE,
    )

    subdivided_faces[:, 0] = mesh.faces.ravel()
    subdivided_faces[:, 1] = mesh.unique_edges().inverse + len(mesh.vertices)
    subdivided_faces[:, 2] = np.repeat(
        np.arange(len(face_centers)) + (len(mesh.vertices) + len(edge_mid_v)),
        4,
        dtype=settings.INT_DTYPE,
    )
    subdivided_faces[:, 3] = (
        subdivided_faces[:, 1].reshape(-1, 4)[:, [3, 0, 1, 2]].ravel()
    )

    if return_dict:
        return {
            "vertices": new_vertices,
            "faces": subdivided_faces,
        }

    else:
        return new_vertices, subdivided_faces


def sorted_unique(connectivity, sorted_=False):
    """Given connectivity array, finds unique entries, based on its axis=1
    sorted values. Returned value will be sorted.

    Parameters
    -----------
    connectivity: (n, d) np.ndarray
    sorted_: bool

    Returns
    --------
    unique_info: Unique2DIntegers
    """
    s_connec = connectivity if sorted_ else np.sort(connectivity, axis=1)

    unique_stuff = arr.unique_rows(
        s_connec,
        return_index=True,
        return_inverse=True,
        return_counts=True,
        dtype_name=settings.INT_DTYPE,
    )

    return helpers.data.Unique2DIntegers(
        unique_stuff[0],  # values
        unique_stuff[1],  # ids
        unique_stuff[2],  # inverse
        unique_stuff[3],  # counts
    )


def _sequentialize_directed_edges(edges, start=None, return_edges=False):
    """
    Sequentialize directed edges.
    """
    # we want to have an np array
    edges = np.asanyarray(edges)

    # Build a lookup_array
    lookup_array = np.full(edges.max() + 1, -1, dtype=settings.INT_DTYPE)
    lookup_array[edges[:, 0]] = edges[:, 1]

    # select starting point - lowest index
    starting_point = int(edges.min()) if start is None else int(start)

    # initialize a set to keep track of processes vertices
    next_candidates = set(edges[:, 0])
    # we want to keep track of single occurrences, as they are line start
    line_starts = set(np.where(np.bincount(edges.ravel()) == 1)[0])
    # for this to work, we can't have a line that starts at column 1.
    # so, we remove those.
    ls_col1 = set(np.where(np.bincount(edges[:, 1].ravel()) == 1)[0])
    line_starts.difference_update(ls_col1)

    polygons = []
    is_polygon = []
    for _ in range(len(edges)):
        # Get polygon - start with first two points
        polygon = [starting_point]
        polygon.append(lookup_array[polygon[-1]])

        # then keep looking until we come to the start.
        # if there's only one edges, this will exit right away
        for _ in range(len(next_candidates)):
            # sequence closes -> polygon
            if polygon[0] == polygon[-1]:
                polygon.pop()
                is_polygon.append(True)
                break

            # sequence leads to unspecified connection -> not polygon
            if lookup_array[polygon[-1]] < 0:
                is_polygon.append(False)
                break
            polygon.append(lookup_array[polygon[-1]])

        polygons.append(polygon)

        # check if we counted all the edges
        # if so, exit
        next_candidates.difference_update(polygons[-1])
        line_starts.difference_update(polygons[-1])

        if len(next_candidates) == len(line_starts) == 0:
            break

        # let's try to find the next starting point
        starting_point = None
        if len(line_starts) != 0:
            starting_point = min(line_starts)
        else:
            starting_point = min(next_candidates)

    if not return_edges:
        return polygons, is_polygon

    else:
        polygon_edges = []
        for p, is_p in zip(polygons, is_polygon):
            polygon_edges.append(sequence_to_edges(p, closed=is_p))

        return polygon_edges, is_polygon


def _sequentialize_edges(edges, start=None, return_edges=False):
    """
    sequentialize undirected edges. No overlaps are allowed, for now.
    """
    edges = np.asanyarray(edges)

    # only applicable to closed polygons and open lines
    # not for arbitrarily connected edges
    bc = np.bincount(edges.ravel())
    if not all(bc < 3):
        raise ValueError(
            "This function currently supports individual lines/polygons "
            "search. Given edges include connections with more than 3 edges."
        )

    # we want to keep track of single occurrences, as they are line start
    line_starts = set(np.where(bc == 1)[0])

    # initialize a set to keep track of processes vertices
    next_candidates = set(edges.ravel())

    # create a look up to each edge column
    edge_col = collections.namedtuple("a", "b")
    edge_col.a = edges[:, 0]
    edge_col.b = edges[:, 1]

    # create trees for each edge column
    tree = collections.namedtuple("a", "b")
    tree.a = napf.KDT(edge_col.a.reshape(-1, 1))
    tree.b = napf.KDT(edge_col.b.reshape(-1, 1))

    # radius search size
    r = 0.1

    current_id = np.argmin(edge_col.a) if start is None else start
    start_value = int(edge_col.a[current_id])
    other_col = edge_col.b

    polygons = []
    is_polygon = []
    for _ in range(len(edges)):
        polygon = [start_value, other_col[current_id]]

        #        while polygon[0] != polygon[-1]:
        for _ in range(len(next_candidates)):
            if polygon[0] == polygon[-1]:
                break
            # search for ids
            a_ids = tree.a.radius_search([[polygon[-1]]], r, True)[0][0]
            b_ids = tree.b.radius_search([[polygon[-1]]], r, True)[0][0]

            # in total, there should be 2 otherwise, we can end this search
            # and this is not a polygon
            hits = len(a_ids) + len(b_ids)
            if hits != 2:
                break

            # so, we have 2 hits. we need to get the partner of non-current_id
            # index
            found = False
            for ai in a_ids:
                if ai != current_id:
                    found = True
                    current_id = ai
                    polygon.append(edge_col.b[current_id])
                    break
            if found:
                continue

            for bi in b_ids:
                if bi != current_id:
                    found = True
                    current_id = bi
                    polygon.append(edge_col.a[current_id])
                    break
            if found:
                continue

            raise RuntimeError(
                "Something went wrong. Please report this issue to "
                "github.com/tataratat/gustaf/issues]"
            )

        # if indices closes itself, it is a polygon.
        # Otherwise it is an open line
        if polygon[0] == polygon[-1]:
            polygon.pop()
            is_polygon.append(True)
        else:
            is_polygon.append(False)
        polygons.append(polygon)

        # check if we counted all the edges
        # if so, exit
        # let's try to find the next starting point
        next_candidates.difference_update(polygons[-1])
        line_starts.difference_update(polygons[-1])

        if len(next_candidates) == len(line_starts) == 0:
            break

        # Assign next starting point
        # generally first value of (those are min() in a set) line start or
        # leftover candidates
        start_value = None
        if len(line_starts) != 0:
            start_value = min(line_starts)
        else:
            start_value = min(next_candidates)

        # adjust state values
        a_ids = tree.a.radius_search([[start_value]], r, True)[0][0]
        if len(a_ids) != 0:
            current_id = a_ids[0]
            other_col = edge_col.b
            continue
        b_ids = tree.b.radius_search([[start_value]], r, True)[0][0]
        if len(b_ids) != 0:
            current_id = b_ids[0]
            other_col = edge_col.a
            continue

        raise RuntimeError(
            "Something went wrong. Please report this issue to "
            "github.com/tataratat/gustaf/issues"
        )

    if not return_edges:
        return polygons, is_polygon

    else:
        polygon_edges = []
        for p, is_p in zip(polygons, is_polygon):
            polygon_edges.append(sequence_to_edges(p, closed=is_p))

        return polygon_edges


def sequentialize_edges(edges, start=None, return_edges=False, directed=False):
    """
    Organize edge connectivities to describe polygon or a line.
    This supports edges that describes separated/individual polygons and lines.
    In other words, it doesn't support edges of overlapping vertices.

    Parameters
    -----------
    edges: (n, 2) list-like
    start: int
      (Optional) Specify starting point. It will take minimum index otherwise.
    return_edges: bool
      (Optional) Default is False. If set True, returns sequences as edges.
    directed: bool
      (Optional) Default is False. Set True, if given edges are directed.
      It should return the result faster.

    Returns
    --------
    sequences: list
      list of vertex ids. Or edges iff return_edges is True.
    is_polygon: list
      Tells if the sequence is a polygon or a line.

    Examples
    ---------
    >>> e = gus.Edges(vertices, edges)
    >>> ordered_sequence, is_polygon = sequentialize_edges(e.edges)

    >>> f = gus.Faces(vertices, faces)
    >>> ordered_sequence, is_polygon = sequentialize_edges(
    ...     f.edges()[f.single_edges()]
    ... )
    """
    if directed:
        return _sequentialize_directed_edges(edges, start, return_edges)
    else:
        return _sequentialize_edges(edges, start, return_edges)
